"""
This module provides a simple implementation of a file-based locking mechanism.

The main class `FileLock` uses OS-level file system operations for locking
and unlocking.
This kind of lock can be used to prevent the simultaneous execution of a piece
of code by different processes.
"""
import os
import time


class FileLockTimeout(Exception):
    """
    Exception raised when a file lock operation fails.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="Timeout occurred."):
        """Initialize the exception."""
        self.message = message
        super().__init__(self.message)


class FileLock:
    """
    Provides a context manager-based file lock mechanism.

    Attributes:
        file_name: The name of the file to be used as the lock.
        timeout (int): The maximum time to wait for the lock to be released.
    """

    def __init__(self, file_name: str, timeout: int = 10) -> None:
        """
        Construct a new FileLock object.

        Args:
            file_name: The name of the file to be used as the lock.
            timeout: The maximum time to wait for the lock to be released.
        """
        self.file_name = file_name
        self.timeout = timeout
        self.fd = None
        self._lock_held = False

    def acquire(self, timeout: int = None, poll_interval: float = 0.05) -> None:
        """
        Acquire the file lock.

        If the lock is currently in use, wait until it is released,
        or until the specified timeout has passed.

        Args:
            timeout: The maximum time to wait for the lock to be released.
                Default is None, which means use self.timeout.
            poll_interval: Time interval to check the lock file's existence.

        Raises:
            FileLockTimeout: If the timeout has passed and the lock
                has not been released.
        """
        timeout = timeout or self.timeout
        start_time = time.time()
        while True:
            try:
                self.fd = os.open(self.file_name, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                self._lock_held = True
                break
            except FileExistsError:
                if time.time() - start_time >= timeout:
                    msg = f"Timeout trying to lock {self.file_name}"
                    raise FileLockTimeout(msg)
                else:
                    time.sleep(poll_interval)

    def release(self, force: bool = False) -> None:
        """
        Release the file lock.

        Args:
            force: If True, force release the lock even if the lock is
                held by others.
        """
        if self._lock_held:
            os.close(self.fd)
            os.unlink(self.file_name)
        elif force:
            try:
                os.unlink(self.file_name)
            except OSError:
                pass
        self._lock_held = False

    @property
    def is_locked(self) -> bool:
        """
        Check if the file lock is locked.

        Returns:
            bool: True if the lock file exists, False otherwise.
        """
        return os.path.exists(self.file_name)

    def __enter__(self) -> "FileLock":
        """Acquire the lock when entering the context."""
        self.acquire()
        return self

    def __exit__(self, type, value, traceback) -> None:
        """Release the lock when exiting the context."""
        self.release()



from unittest.mock import patch

import pytest

from inet_nm._helpers import (
    inquirer,
    nm_prompt_choice,
    nm_prompt_confirm,
    nm_prompt_input,
)


def test_nm_prompt_choice_single_choice():
    """Test nm_prompt_choice function with single choice."""
    assert nm_prompt_choice("Choose", ["choice1"]) == "choice1"


def test_nm_prompt_choice_multiple_choices():
    """Test nm_prompt_choice function with multiple choices."""
    with patch("inet_nm._helpers._prompt_choice", return_value="choice2"):
        assert nm_prompt_choice("Choose", ["choice1", "choice2"]) == "choice2"


@pytest.mark.skipif(not inquirer, reason="inquirer not installed")
def test_nm_prompt_choice_multiple_choices_inquirer():
    """Test nm_prompt_choice function with multiple choices mocking inquirer."""
    with patch("inquirer.prompt", return_value={"res": "choice2"}):
        assert nm_prompt_choice("Choose", ["choice1", "choice2"]) == "choice2"


def test_nm_prompt_choice_with_custom():
    """Test nm_prompt_choice function with custom input."""
    with patch("inet_nm._helpers._prompt_choice", return_value="custom"), patch(
        "builtins.input", return_value="mychoice"
    ):
        assert (
            nm_prompt_choice("Choose", ["choice1", "choice2"], custom=True)
            == "mychoice"
        )


def test_nm_prompt_choice_multiple_choices_func():
    """Test nm_prompt_choice function with multiple choices non-string choices."""

    def _test(val: int):
        msg = f"val=={val+1}"
        return msg

    with patch("inet_nm._helpers._prompt_choice", return_value="val==2"):
        assert nm_prompt_choice("Choose", [1, 10], _test) == 1


def test_nm_prompt_input():
    """Test nm_input function."""
    with patch("builtins.input", return_value="myinput"):
        assert nm_prompt_input("Enter value") == "myinput"


@pytest.mark.skipif(not inquirer, reason="inquirer not installed")
def test_nm_prompt_confirm():
    """Test nm_prompt_confirm function."""
    with patch("inquirer.prompt", return_value={"res": True}):
        assert nm_prompt_confirm("Confirm") is True
    with patch("inquirer.prompt", return_value={"res": False}):
        assert nm_prompt_confirm("Confirm") is False



from threading import Thread
from typing import Dict
from unittest.mock import patch

import pytest

from inet_nm.data_types import NmNode
from inet_nm.runner_base import NmNodesRunner


class MockNmNodesRunner(NmNodesRunner):
    def func(self, node: NmNode, idx: int, env: Dict[str, str]):
        return True


@pytest.fixture
def dummy_nodes():
    node1 = NmNode(
        serial="1",
        vendor="vendor1",
        product_id="product1",
        vendor_id="vendor_id1",
        model="model1",
        driver="driver1",
        board="board1",
    )
    node2 = NmNode(
        serial="2",
        vendor="vendor2",
        product_id="product2",
        vendor_id="vendor_id2",
        model="model2",
        driver="driver2",
        board="board2",
    )
    return [node1, node2]


def test_acquire_and_release(dummy_nodes):
    runner = MockNmNodesRunner(nodes=dummy_nodes)

    with patch.object(runner.locks[0], "acquire") as mock_acquire:
        with patch.object(runner.locks[0], "release") as mock_release:
            runner.acquire()
            mock_acquire.assert_called_once()

            runner.release()
            mock_release.assert_called_once()


def test_run_without_acquire(dummy_nodes):
    runner = MockNmNodesRunner(nodes=dummy_nodes)

    with pytest.raises(Exception):
        runner.run()


def test_run_with_acquire(dummy_nodes):
    runner = MockNmNodesRunner(nodes=dummy_nodes)
    runner.acquire()

    with patch.object(Thread, "start") as mock_start, patch.object(
        Thread, "join"
    ) as mock_join:
        runner.run()
        assert mock_start.call_count == len(dummy_nodes)
        assert mock_join.call_count == len(dummy_nodes)



"""
This module includes classes that are used to run applications on nodes.

Classes:
    NmShellRunner: Runs shell commands on nodes.
    NmTmuxPanedRunner: Runs tmux with paned windows.
    NmTmuxWindowedRunner: Runs tmux with individual windows for each node.
"""

import os
import subprocess
import time
from typing import Dict

from inet_nm.data_types import NmNode
from inet_nm.runner_base import NmNodesRunner


class NmShellRunner(NmNodesRunner):
    """Runs shell commands on nodes.

    This class inherits from NmNodesRunner and overrides the func method
    to execute shell commands on nodes.
    """

    cmd = "echo $NM_IDX"

    def func(self, node: NmNode, idx: int, env: Dict[str, str]):
        """Execute shell commands on nodes.

        Args:
            node: Node to run the command on.
            idx: Index of the node.
            env: Environment variables for the command.
        """
        full_env = {**os.environ, **env}  # Merge original and new environment variables
        full_env = {
            k: str(v) for k, v in full_env.items()
        }  # Cast everything in env to a string
        subprocess.run(self.cmd, shell=True, env=full_env)


class NmTmuxBaseRunner(NmNodesRunner):
    """Base class for tmux runners.

    This class inherits from NmNodesRunner and sets up a tmux session.
    """

    session_name: str = "default"
    cmd: str = None
    SETUP_WAIT = 0.2

    def post(self):
        """Run after the operations on nodes have completed.

        It attaches to the tmux session and keeps checking if
        the session is still active.
        """
        os.system(f"tmux attach -t {self.session_name}")

        while True:
            result = subprocess.run(
                f"tmux has-session -t {self.session_name}",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            if result.returncode != 0:
                break


class NmTmuxPanedRunner(NmTmuxBaseRunner):
    """Run tmux with paned windows.

    This class inherits from NmTmuxBaseRunner and sets up a tmux
    session with panes.
    """

    def func(self, node: NmNode, idx: int, env: Dict[str, str]):
        """Set up a tmux session with paned windows.

        Args:
            node: Node to run the command on.
            idx: Index of the node.
            env: Environment variables for the command.
        """
        time.sleep(idx * self.SETUP_WAIT)

        e_args = " -e " + " -e ".join([f"{key}={value}" for key, value in env.items()])
        if idx != 0:
            subprocess.run(
                f"tmux split-window {e_args} -t {self.session_name}", shell=True
            )
        else:
            subprocess.run(f"tmux new-session -d -s {self.session_name}", shell=True)
            subprocess.run(
                f"tmux respawn-window -k {e_args} -t {self.session_name} -t 0.{idx}",
                shell=True,
            )

        # Select pane
        subprocess.run(f"tmux select-pane -t {idx}", shell=True)

        # Execute command
        if self.cmd:
            subprocess.run(
                f"tmux send-keys -t {self.session_name}:0.{idx} '{self.cmd}' Enter",
                shell=True,
            )

        # Evenly distribute panes
        subprocess.run(f"tmux select-layout -t {self.session_name} tiled", shell=True)


class NmTmuxWindowedRunner(NmTmuxBaseRunner):
    """Runs tmux with individual windows for each node.

    This class inherits from NmTmuxBaseRunner and sets up a tmux session with
    individual windows for each node.
    """

    def func(self, node: NmNode, idx: int, env: Dict[str, str]):
        """Sets up a tmux session with individual windows for each node.

        Args:
            node: Node to run the command on.
            idx: Index of the node.
            env: Environment variables for the command.
        """
        env = {k: str(v) for k, v in env.items()}  # Cast everything in env to a string
        uid = env["NM_UID"]
        session_name = self.session_name

        time.sleep(idx * self.SETUP_WAIT)

        e_args = " -e " + " -e ".join([f"{key}={value}" for key, value in env.items()])
        if idx != 0:
            subprocess.run(
                f"tmux new-window -t {session_name}:{idx} {e_args}", shell=True
            )
        else:
            subprocess.run(f"tmux new-session -d -s {session_name}", shell=True)
            subprocess.run(
                f"tmux respawn-window -k {e_args} -t {session_name} -t 0.{idx}",
                shell=True,
            )

        subprocess.run(f"tmux rename-window -t {session_name}:{idx} {uid}", shell=True)

        if self.cmd:
            subprocess.run(
                f"tmux send-keys -t {session_name}:{idx} '{self.cmd}' Enter", shell=True
            )



"""Interactive user prompts and general basic helpers."""

import os
import readline
import sys
from typing import Callable, List, Optional

try:
    import inquirer
except ImportError:
    inquirer = None


def nm_prompt_choice(
    msg: str,
    choices: List[str],
    choice_func: Optional[Callable] = None,
    custom: bool = False,
) -> str:
    """
    Interactive user prompt to choose from a list of choices.

    If inquirer is not installed, a simpler choice will be used.
    If choice_func is provided, it will be applied to the choices for display.

    Args:
        msg: Message to display to the user.
        choices: List of choices.
        choice_func: Function to format choices for display.
        custom: Allow user to provide custom input.

    Returns:
        str: The selected or entered choice.

    Raises:
        ValueError: If no choices are available.
    """
    if len(choices) == 0:
        raise ValueError(f"No choices available for: {msg}")
    if len(choices) == 1:
        return choices[0]
    if custom:
        choices.append("custom")
    if choice_func:
        _choices = [choice_func(choice) for choice in choices]
    else:
        _choices = choices

    res = _prompt_choice(msg, _choices)
    if res == "custom":
        return input("Custom value: ")
    return choices[_choices.index(res)] if choice_func else res


def _prompt_choice(
    msg: str,
    choices: List[str],
) -> str:
    if inquirer:
        return _inquirer_prompt_choice(msg, choices)
    return _basic_prompt_choice(msg, choices)


def _inquirer_prompt_choice(
    msg: str,
    choices: List[str],
) -> str:
    res = inquirer.prompt(
        [inquirer.List("res", message=msg, choices=choices)],
        raise_keyboard_interrupt=True,
    )
    return res["res"]


def _basic_prompt_choice(msg: str, choices: List[str]) -> str:
    nm_print(msg)
    for i, choice in enumerate(choices):
        nm_print(f"{i+1}. {choice}")
    while True:
        try:
            choice = int(input("> "))
            if choice < 1 or choice > len(choices):
                raise ValueError
            break
        except ValueError:
            nm_print(
                "Invalid choice. Please enter a number between 1 and " f"{len(choices)}"
            )
    return choices[choice - 1]


class _Completer:
    """Helper class for command line autocompletion."""

    def __init__(self, options: List[str]):
        """
        Initialize _Completer with a list of options.

        Args:
            options: List of autocompletion options.
        """
        self.options = sorted(options)

    def complete(self, text: str, state: int) -> Optional[str]:
        """
        Autocomplete function used by readline.

        Args:
            text: Partial text to complete.
            state: Index of the match being requested.

        Returns:
            Match at index 'state', or None if index is out
                of bounds.
        """
        if state == 0:  # on first trigger, build possible matches
            if not text:
                self.matches = self.options[:]
            else:
                mat = [s for s in self.options if s and s.startswith(text)]
                self.matches = mat

        # return match indexed by state
        try:
            return self.matches[state]
        except IndexError:
            return None

    def display_matches(self, substitution: str, matches: List[str], _):
        """
        Display possible matches to the user.

        Args:
            substitution (str): The substitution made so far.
            matches (List[str]): List of matches found.
        """
        line_buffer = readline.get_line_buffer()
        columns = os.environ.get("COLUMNS", 80)

        nm_print()
        full_matches = [substitution + match for match in matches]

        tpl = "{:<" + str(int(max(map(len, full_matches)) * 1.2)) + "}"

        buffer = ""
        for match in full_matches:
            match = tpl.format(match[len(substitution) :])
            if len(buffer + match) > columns:
                nm_print(buffer)
                buffer = ""
            buffer += match

        if buffer:
            nm_print(buffer)

        nm_print("> ", end="")
        nm_print(line_buffer, end="")
        sys.stdout.flush()


def nm_prompt_input(prompt: str, options: List[str] = None) -> str:
    """
    Interactive user prompt with auto-completion.

    Args:
        prompt: Message to display to the user.
        options: List of auto-completion options.

    Returns:
        str: User input.
    """
    if options is None:
        options = []
    completer = _Completer(list(set(options)))
    readline.set_completer_delims(" \t\n;")
    readline.set_completer(completer.complete)
    readline.parse_and_bind("tab: complete")
    readline.set_completion_display_matches_hook(completer.display_matches)

    nm_print(prompt)
    return nm_input("> ")


def nm_prompt_confirm(msg: str, default: bool = False) -> bool:
    """
    Interactive user prompt for confirmation.

    Args:
        msg: Message to display to the user.
        default: Default value.

    Returns:
        bool: User confirmation.
    """
    if inquirer:
        res = inquirer.prompt(
            [inquirer.Confirm("res", message=msg, default=default)],
            raise_keyboard_interrupt=True,
        )
        return res["res"]
    else:
        return nm_input(f"{msg} [y/N] ").lower() == "y"


def nm_print(*args, **kwargs):
    """Abstraction of print."""
    print(*args, **kwargs)


def nm_input(*args, **kwargs):
    """Abstraction of input."""
    return input(*args, **kwargs)



"""Pytest module for testing all runner apps."""
from typing import List
from unittest.mock import patch

import pytest

from inet_nm.data_types import NmNode
from inet_nm.locking import get_locked_uids
from inet_nm.runner_apps import NmShellRunner, NmTmuxPanedRunner, NmTmuxWindowedRunner


def _nodes(amount=1) -> List[NmNode]:
    nodes = []
    for i in range(1, amount + 1):
        nodes.append(
            NmNode(
                serial=f"serial{i}",
                vendor_id="vendor_id",
                product_id="product_id",
                vendor="vendor",
                model="model",
                driver="driver",
                board="board",
                features_provided=["feature0", f"feature{i}"],
            )
        )
    return nodes


def test_NmShellRunner():
    """Test default script running on dummy nodes."""
    nodes = _nodes(2)
    with NmShellRunner(nodes) as runner:
        runner.run()
    uids = get_locked_uids()
    assert nodes[0].uid not in uids
    assert nodes[1].uid not in uids


@pytest.mark.parametrize("tmux_runner", [NmTmuxPanedRunner, NmTmuxWindowedRunner])
def test_NmTmuxRunner(tmux_runner):
    """Mock tmux calls."""
    node = _nodes(1)[0]
    with patch("subprocess.run"):
        with tmux_runner([node]) as runner:
            runner.run()
    uids = get_locked_uids()
    assert node.uid not in uids



import argparse
import os
import subprocess
from pathlib import Path
from typing import List

import inet_nm.config as cfg


def _get_names(cmd: str, cwd: Path) -> List[str]:
    """Call make info-boards and parse the output to a list."""
    res = subprocess.run(cmd.split(), cwd=cwd, capture_output=True)
    if res.returncode != 0:
        raise RuntimeError(f"Failed to run {cmd}")
    return res.stdout.decode().split()


def _get_features_provided(
    cmd: str, cwd: Path, board_env_var: str, board_names: List[str]
):
    boards = {}
    errors = []
    for bn in board_names:
        print(f"Getting features_provided for {bn}")
        env = os.environ.copy()
        env[board_env_var] = bn
        res = subprocess.run(cmd.split(), cwd=cwd, env=env, capture_output=True)
        if res.returncode != 0:
            errors.append(bn)
            print(f"FAILED to run {cmd} for {bn}")
        boards[bn] = res.stdout.decode().split()
    if errors:
        raise RuntimeError(f"Failed to run {cmd} for {errors}")
    return boards


def cli_update_boards_from_os():
    parser = argparse.ArgumentParser(description="Cache a list of boards")
    cfg.config_arg(parser)
    parser.add_argument("basedir", type=Path, help="Path to the board path directory")
    parser.add_argument(
        "-i",
        "--board-info",
        type=str,
        default="make info-boards",
        help="Command to get board info, defaults to 'make info-boards'",
    )
    parser.add_argument(
        "-f",
        "--board-features",
        type=str,
        default="make info-features-provided",
        help="Command to get board features, defaults to 'make info-features-provided'",
    )
    parser.add_argument(
        "-n",
        "--board-env-var",
        type=str,
        default="BOARD",
        help="The env var to indicate the board name for features provided,"
        "defaults to 'BOARD'",
    )

    args = parser.parse_args()

    try:
        bns = _get_names(args.board_info, args.basedir)
        board_info = _get_features_provided(
            args.board_features, args.basedir, args.board_env_var, bns
        )
        cfg.save_riot_board_info(args.config, board_info)
        print(f"\nUpdated {cfg.board_info_path(args.config)}")
    except KeyboardInterrupt:
        print("\nUser aborted")
    except RuntimeError as e:
        print(f"\nError: {e}")



import argparse

import inet_nm.check as check
import inet_nm.config as cfg
import inet_nm.runner_apps as apps


def nm_tmux():
    parser = argparse.ArgumentParser(description="Commission USB boards")
    parser.add_argument(
        "-w", "--window", action="store_true", help="Open a window for each node."
    )
    parser.add_argument(
        "-t",
        "--timeout",
        type=float,
        default=None,
        help="Wait until node available in seconds.",
    )
    parser.add_argument(
        "-x",
        "--cmd",
        type=str,
        default=None,
        help="Command to send after starting tmux session.",
    )
    cfg.config_arg(parser)
    check.check_args(parser)
    args = parser.parse_args()
    kwargs = vars(args)
    window = kwargs.pop("window")
    timeout = kwargs.pop("timeout")
    cmd = kwargs.pop("cmd")
    nodes = check.get_filtered_nodes(**kwargs)
    if len(nodes) == 0:
        print("No available nodes!")
        return
    if window:
        with apps.NmTmuxWindowedRunner(nodes, default_timeout=timeout) as runner:
            runner.cmd = cmd
            runner.run()
    else:
        with apps.NmTmuxPanedRunner(nodes, default_timeout=timeout) as runner:
            runner.cmd = cmd
            runner.run()


def nm_exec():
    parser = argparse.ArgumentParser(description="Commission USB boards")
    parser.add_argument(
        "cmd", type=str, help="Command to send after starting tmux session."
    )
    parser.add_argument(
        "-t",
        "--timeout",
        type=float,
        default=None,
        help="Wait until node available in seconds.",
    )

    cfg.config_arg(parser)
    check.check_args(parser)
    args = parser.parse_args()
    kwargs = vars(args)
    timeout = kwargs.pop("timeout")
    cmd = kwargs.pop("cmd")
    nodes = check.get_filtered_nodes(**kwargs)
    if len(nodes) == 0:
        print("No available nodes!")
        return
    with apps.NmShellRunner(nodes, default_timeout=timeout) as runner:
        runner.cmd = cmd
        runner.run()



import pytest

from inet_nm.config import (
    get_lock_path,
    load_nodes,
    load_riot_board_info,
    save_nodes,
    save_riot_board_info,
)
from inet_nm.data_types import NmNode


def test_save_load_riot_board_info(tmp_path):
    """Test saving and loading the riot board info."""
    board_info = {"board": "riot", "version": "1.0"}
    save_riot_board_info(tmp_path, board_info)
    loaded_board_info = load_riot_board_info(tmp_path)
    assert loaded_board_info == board_info
    assert len(load_riot_board_info("does_not_exist")) == 0


@pytest.mark.parametrize(
    "nodes_data",
    [
        [],
        [
            {
                "serial": "1",
                "vendor_id": "vendor_id1",
                "product_id": "product1",
                "vendor": "vendor1",
                "model": "model1",
                "driver": "driver1",
            }
        ],
    ],
)
def test_save_load_nodes(tmp_path, nodes_data):
    """Test saving and loading the nodes."""
    nodes = [NmNode.from_dict(data) for data in nodes_data]
    save_nodes(tmp_path, nodes)
    loaded_nodes = load_nodes(tmp_path)
    assert loaded_nodes == nodes
    assert len(load_nodes("does_not_exist")) == 0


def test_get_lock_path(tmp_path):
    """Test getting the lock path."""
    lock_path = get_lock_path(tmp_path)
    assert lock_path == tmp_path / "locks"



from copy import copy
from unittest.mock import MagicMock, patch

import pytest

import inet_nm.commissioner as commissioner
from inet_nm.data_types import NmNode


@pytest.fixture
def dummy_nodes():
    """Return a list of dummy nodes."""
    node1 = NmNode(
        serial="1",
        vendor="vendor1",
        product_id="product1",
        vendor_id="vendor_id1",
        model="model1",
        driver="driver1",
    )
    node2 = NmNode(
        serial="2",
        vendor="vendor2",
        product_id="product2",
        vendor_id="vendor_id2",
        model=None,
        driver="driver2",
    )
    return [node1, node2]


@pytest.fixture
def mock_devices():
    # Mock the devices returned by context.list_devices
    mock_device_1 = MagicMock()
    mock_device_1.device_node = "/dev/ttyUSB0"
    mock_device_1.find_parent.return_value.get.side_effect = [
        "vendor_id1",
        "product1",
        "1",
        "vendor1",
        "model1",
        "driver1",
    ]

    mock_device_2 = MagicMock()
    mock_device_2.device_node = "/dev/ttyUSB1"
    mock_device_2.find_parent.return_value.get.side_effect = [
        "vendor_id2",
        "product2",
        "2",
        "vendor2",
        None,
        "driver2",
    ]

    return [mock_device_1, mock_device_2]


def test_get_devices_from_tty(mock_devices, dummy_nodes):
    """Mock devices returned by context.list_devices."""
    # Mock the context.list_devices method
    list_devices_mock = MagicMock()
    list_devices_mock.return_value = mock_devices

    with patch("pyudev.Context.list_devices", list_devices_mock):
        assert commissioner.get_devices_from_tty() == dummy_nodes


def test_get_tty_from_nm_node(mock_devices, dummy_nodes):
    """Test if the tty can be checked fro a given node."""
    with patch("pyudev.Context.list_devices", MagicMock(return_value=mock_devices[:1])):
        assert (
            commissioner.get_tty_from_nm_node(dummy_nodes[0])
            == mock_devices[0].device_node
        )

        with pytest.raises(commissioner.TtyNotPresent):
            commissioner.get_tty_from_nm_node(dummy_nodes[1])


def test_add_node_to_nodes(dummy_nodes):
    """Test if a node can be added to a list of nodes."""
    nodes = copy(dummy_nodes)
    new_node = NmNode(
        vendor_id="vendor_id3",
        product_id="product3",
        serial="3",
        vendor="vendor3",
        model="model3",
        driver="driver3",
    )
    assert len(commissioner.add_node_to_nodes(nodes, new_node)) == 3
    assert len(commissioner.add_node_to_nodes(nodes, nodes[-1])) == 2

    different_old_node = copy(nodes[-1])
    different_old_node.model = "something different"
    new_nodes = commissioner.add_node_to_nodes(nodes, different_old_node)
    assert len(new_nodes) == 2
    assert new_nodes[-1].model == "something different"



"""
Runs applications on nodes.

This module provides the NmNodesRunner class which handles running
operations on multiple nodes.
An operation is defined by a method `func` which is to be implemented
in the subclass.
"""
from threading import Thread
from typing import Dict, List

import inet_nm.locking as lk
from inet_nm.commissioner import TtyNotPresent, get_tty_from_nm_node
from inet_nm.data_types import NmNode, NodeEnv
from inet_nm.filelock import FileLock


class NmNodesRunner:
    """
    Class to handle running operations on nodes.

    The class manages locking/unlocking nodes, running operations on each node
    concurrently in separate threads, and handles cleanup after operations
    have completed.

    The operations to be run are defined by a method `func` which is to
    be implemented in the subclass.
    """

    def __init__(self, nodes: List[NmNode], default_timeout: int = None):
        """
        Initialize a new instance of NmNodesRunner.

        Args:
            nodes: A list of NmNode instances to be managed.
            default_timeout: Default timeout value for file lock acquisition.
        """
        self.nodes = nodes
        self.default_timeout = default_timeout
        self.lockable_nodes = [
            (node, FileLock(lk.get_lock_path(node), timeout=default_timeout))
            for node in nodes
        ]
        self.locks = [lock for _, lock in self.lockable_nodes]
        self._acquired = False

    def pre(self):
        """Override in the subclass if pre-operation steps are needed."""
        pass

    def post(self):
        """Override in the subclass if post-operation steps are needed."""
        pass

    def func(self, node: NmNode, idx: int, env: Dict[str, str]):
        """
        Function to execute.

        It must be implemented in the subclass.

        Args:
            node: The node to run the function on.
            idx: The index of the node in the list of nodes.
            env: A dictionary of environment variables.
        """
        raise NotImplementedError("You must implement a func() method")

    def acquire(self, timeout: float = None):
        """
        Acquire file locks for all nodes.

        This method must be called before running operations on nodes.

        Args:
            timeout (float): Timeout value for file lock acquisition.
                If None, default_timeout is used.
        """
        for lock in self.locks:
            lock.acquire(timeout=timeout or self.default_timeout)
        self._acquired = True

    def release(self):
        """Release all acquired file locks."""
        for lock in self.locks:
            lock.release()
        self._acquired = False

    def run(self):
        """
        Run operations on all nodes.

        This method spawns a new thread for each node and runs the
        operation concurrently on all nodes.
        The operation to run is defined by the `func` method.
        """
        if not self._acquired:
            raise Exception("You must call acquire() before calling run()")

        self.pre()

        self.threads = []
        for idx, node in enumerate(self.nodes):
            try:
                nm_port = get_tty_from_nm_node(node)
            except TtyNotPresent:
                nm_port = "Unknown"
            node_env = NodeEnv(
                NM_IDX=idx,
                NM_UID=node.uid,
                NM_SERIAL=node.serial,
                NM_BOARD=node.board,
                NM_PORT=nm_port,
            ).to_dict()

            thread = Thread(target=self.func, args=(node, idx, node_env))
            thread.start()
            self.threads.append(thread)

        for thread in self.threads:
            thread.join()

        self.post()
        self.release()

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        self.release()
        self._acquired = False



from pathlib import Path

import pytest

from inet_nm.data_types import NmNode
from inet_nm.filelock import FileLock
from inet_nm.locking import get_lock_path, get_locked_uids, locks_dir, release_all_locks


@pytest.fixture
def dummy_nodes():
    """Return a list of dummy nodes."""
    node1 = NmNode(
        serial="1",
        vendor="vendor1",
        product_id="product1",
        vendor_id="vendor_id1",
        model="model1",
        driver="driver1",
    )
    node2 = NmNode(
        serial="2",
        vendor="vendor2",
        product_id="product2",
        vendor_id="vendor_id2",
        model=None,
        driver="driver2",
    )
    return [node1, node2]


def setup_function():
    nodes = [
        NmNode(
            serial="1",
            vendor="vendor1",
            product_id="product1",
            vendor_id="vendor_id1",
            model="model1",
            driver="driver1",
        ),
        NmNode(
            serial="2",
            vendor="vendor2",
            product_id="product2",
            vendor_id="vendor_id2",
            model=None,
            driver="driver2",
        ),
    ]

    FileLock(str(get_lock_path(nodes[0]))).acquire()
    FileLock(str(get_lock_path(nodes[1]))).acquire()


def teardown_function():
    release_all_locks()


def test_locks_dir():
    """Checks that the dir is a dir..."""
    dir_path = locks_dir()
    assert dir_path.is_dir()


def test_get_locked_uids():
    """Verify the uids are corrected in the config dir."""
    locked_uids = get_locked_uids()
    assert isinstance(locked_uids, list)
    assert len(locked_uids) == 2
    assert "460e1ae2f9c8315c74015d608e5204a6" in locked_uids
    assert "5de815ec154c80e5af83781cf36acf21" in locked_uids


def test_get_lock_path(dummy_nodes):
    """Verify lock path is correct."""
    node = dummy_nodes[0]
    lock_path = get_lock_path(node)
    assert isinstance(lock_path, Path)
    assert lock_path.name == f"{node.uid}.lock"


def test_release_all_locks():
    """Test if all locks released."""
    locked_uids = get_locked_uids()
    assert len(locked_uids) != 0
    release_all_locks()
    locked_uids = get_locked_uids()
    assert len(locked_uids) == 0



"""
This module contains functions for checking the state of the boards.

This is meant for evaluating inventory.
"""
import argparse
import json
from typing import Dict, List

import pyudev

import inet_nm.config as cfg
from inet_nm.data_types import NmNode
from inet_nm.locking import get_locked_uids


def get_connected_uids() -> List[str]:
    """
    Get the UIDs of all connected USB devices.

    Returns:
        A list of UIDs of all connected USB devices.
    """
    uids = []
    context = pyudev.Context()
    for device in context.list_devices(subsystem="tty"):
        parent = device.find_parent("usb", "usb_device")
        if parent is None:
            continue
        vendor_id = parent.get("ID_VENDOR_ID")
        model_id = parent.get("ID_MODEL_ID")
        serial_short = parent.get("ID_SERIAL_SHORT")
        uids.append(NmNode.calculate_uid(model_id, vendor_id, serial_short))
    return uids


def get_nodes_with_state(nodes: List[NmNode], connected=True) -> List[NmNode]:
    """
    Get a list of nodes that are connected or not connected.

    Args:
        nodes: A list of nodes to filter.
        connected: If True, return the nodes that are
            connected. If False, return the nodes that are not connected.

    Returns:
        A list of filtered nodes.
    """
    selected_nodes = []
    connected_uids = get_connected_uids()
    for node in nodes:
        if node.uid in connected_uids:
            if connected:
                selected_nodes.append(node)
        elif not connected:
            selected_nodes.append(node)
    return selected_nodes


def remove_used_nodes(nodes: List[NmNode], used_uids: List[str]) -> List[NmNode]:
    """
    Remove the nodes that are already used from the list of nodes.

    Args:
        nodes: A list of nodes to filter.
        used_uids: A list of UIDs of nodes that are already used.

    Returns:
        List[NmNode]: A list of nodes that are not already used.
    """
    if used_uids is None:
        used_uids = []
    selected_nodes = []
    for node in nodes:
        if node.uid not in used_uids:
            selected_nodes.append(node)
    return selected_nodes


def filter_nodes(nodes: List[NmNode], must_contain: List[str]) -> List[NmNode]:
    """
    Filter the nodes based on the provided features.

    Args:
        nodes: A list of nodes to filter.
        must_contain: A list of features. Only the nodes that contain all of
            these features will be returned.

    Returns:
        A list of filtered nodes.
    """
    selected_nodes = []
    for node in nodes:
        if all([feature in node.features_provided for feature in must_contain]):
            selected_nodes.append(node)
    return selected_nodes


def nodes_to_boards(nodes: List[NmNode]) -> Dict[str, int]:
    """
    Convert a list of nodes to a dictionary of board counts.

    Args:
        nodes: A list of nodes.

    Returns:
        A dictionary where the keys are the names of the boards and the values
            are the counts of those boards.
    """
    boards = {}
    for node in nodes:
        if node.board not in boards:
            boards[node.board] = 0
        boards[node.board] += 1
    return boards


def get_all_features(nodes: List[NmNode]) -> List[str]:
    """
    Get a list of all features provided by the nodes.

    Args:
        nodes: A list of nodes.

    Returns:
        A list of all features.
    """
    features = []
    for node in nodes:
        features.extend(node.features_provided)
    return set(features)


def eval_features(nodes: List[NmNode], features: set, eval_func):
    """
    Evaluate features of nodes using a provided function.

    Args:
        nodes: A list of nodes to evaluate.
        features: A set of features to evaluate.
        eval_func: The function to use to evaluate the features.

    Returns:
        A list of nodes that meet the criteria of the evaluation function.

    Raises:
        ValueError: If the evaluation function cannot be evaluated.
    """
    selected_nodes = []
    try:
        for node in nodes:
            selected_features = {
                feature: feature in node.features_provided for feature in features
            }
            if eval(eval_func, selected_features):
                selected_nodes.append(node)
    except Exception as e:
        raise ValueError(
            f"Could not evaluate {eval_func} with {sorted(features)}"
        ) from e
    return selected_nodes


def skip_duplicate_boards(nodes: List[NmNode]) -> List[NmNode]:
    """
    Remove duplicate boards from a list of nodes.

    Args:
        nodes: A list of nodes.

    Returns:
        A list of nodes without duplicates.
    """
    selected_nodes = []
    boards = set()
    for node in nodes:
        if node.board not in boards:
            selected_nodes.append(node)
        boards.add(node.board)
    return selected_nodes


def check_nodes(
    nodes: List[NmNode],
    all_nodes: bool = False,
    missing: bool = False,
    feat_filter: List[str] = None,
    feat_eval: str = None,
    used: bool = False,
    skip_dups: bool = False,
    locked_nodes: List[str] = None,
) -> List[NmNode]:
    """
    Get a filtered list of nodes based on the provided parameters.

    Args:
        nodes: A list of nodes to check.
        all_nodes: If True, all nodes will be returned.
        missing: If True, only the nodes that are missing will be returned.
        feat_filter: A list of features. Only the nodes that contain all of
            these features will be returned.
        feat_eval: The function to use to evaluate the features.
        used: If True, used nodes will also be returned.
        skip_dups: If True, duplicate nodes will be removed.
        locked_nodes: A list of UIDs of nodes that are locked.

    Returns:
        A list of filtered nodes.
    """
    features = get_all_features(nodes)
    if not all_nodes:
        nodes = get_nodes_with_state(nodes, connected=not missing)
    if not used:
        nodes = remove_used_nodes(nodes, locked_nodes)
    if feat_filter:
        nodes = filter_nodes(nodes, feat_filter)
    if feat_eval:
        nodes = eval_features(nodes, features, feat_eval)
    if skip_dups:
        nodes = skip_duplicate_boards(nodes)
    return nodes


def check_args(parser: argparse.ArgumentParser):
    """
    Define the arguments for the argparse parser.

    Args:
        parser: The argparse parser.
    """
    parser.add_argument(
        "-f",
        "--feat-filter",
        nargs="+",
        help="Filter all boards that don't provide these features",
    )
    parser.add_argument(
        "-a",
        "--all-nodes",
        action="store_true",
        help="Show all boards, regardless of connection",
    )
    parser.add_argument(
        "-m", "--missing", action="store_true", help="Show all missing boards"
    )
    parser.add_argument(
        "-e", "--feat-eval", type=str, help="Evaluate features with this function"
    )
    parser.add_argument(
        "-u", "--used", action="store_true", help="Show used boards as well"
    )
    parser.add_argument(
        "-s", "--skip-dups", action="store_true", help="Skip duplicate boards"
    )


def get_filtered_nodes(
    config: str,
    all_nodes: bool = False,
    missing: bool = False,
    feat_filter: List[str] = None,
    feat_eval: str = None,
    used: bool = False,
    skip_dups: bool = False,
) -> List[NmNode]:
    """
    Get a list of nodes based on the provided parameters.

    Args:
        config: The configuration path for the nodes.
        all_nodes: If True, all nodes will be returned.
        missing: If True, only the nodes that are missing will be returned.
        feat_filter: A list of features. Only the nodes that contain all of
            these features will be returned.
        feat_eval: The function to use to evaluate the features.
        used: If True, used nodes will also be returned.
        skip_dups: If True, duplicate nodes will be removed.

    Returns:
        A list of filtered nodes.
    """
    nodes = cfg.load_nodes(config)
    locked_nodes = get_locked_uids()
    nodes = check_nodes(
        nodes, all_nodes, missing, feat_filter, feat_eval, used, skip_dups, locked_nodes
    )
    return nodes


def all_features_from_nodes(nodes: List[NmNode]) -> List[str]:
    """
    Get a list of all features provided by the nodes.

    Args:
        nodes: A list of nodes.

    Returns:
        A list of all features.
    """
    features = []
    for node in nodes:
        features.extend(node.features_provided)
    return sorted(list(set(features)))


def cli_count_board_inventory():
    """CLI entrypoint for counting the inventory of boards."""
    parser = argparse.ArgumentParser(description="Check the state of the boards")
    cfg.config_arg(parser)
    check_args(parser)
    parser.add_argument(
        "--show-features", action="store_true", help="Shows all features for all boards"
    )
    args = parser.parse_args()
    kwargs = vars(args)
    show_features = kwargs.pop("show_features")

    nodes = get_filtered_nodes(**kwargs)

    if show_features:
        info = all_features_from_nodes(nodes)
    else:
        info = nodes_to_boards(nodes)
    out = json.dumps(info, indent=2, sort_keys=True)
    print(out)



import random

import pytest

from inet_nm.filelock import FileLock, FileLockTimeout


def test_filelock(tmpdir):
    # Create a lock file path in the platform-independent temporary directory
    lock_file = tmpdir.join(f"testlock-{random.randint(0, 100000)}.lock")

    with FileLock(str(lock_file), timeout=0.1) as lock:
        assert lock.is_locked
        with pytest.raises(FileLockTimeout):
            with FileLock(str(lock_file), timeout=0.1):
                pass

    lock = FileLock(str(lock_file), timeout=1)
    assert not lock.is_locked



"""Provide utilities for interacting with the inet_nm configuration.

It includes functionality for saving and loading information about
the boards and nodes, as well as for handling the path configuration.
"""

import argparse
import json
import os
from pathlib import Path
from typing import Dict, List, Union

from inet_nm.data_types import NmNode


def _file_from_path(config_dir: Union[Path, str], file_name: str) -> Path:
    """
    Return a file path constructed from config_dir and file_name.

    Args:
        config_dir: Directory for the configuration files.
        file_name: Name of the file.

    Returns:
        The full path to the file.
    """
    config_dir = Path(config_dir).expanduser()
    file_path = config_dir / file_name
    config_dir.parent.mkdir(parents=True, exist_ok=True)
    return file_path


def nodes_path(config_dir: Union[Path, str]) -> Path:
    """
    Get the path to the nodes file.

    Args:
        config_dir: Directory for the configuration files.

    Returns:
        The full path to the nodes file.
    """
    return _file_from_path(config_dir, "nodes.json")


def board_info_path(config_dir: Union[Path, str]) -> Path:
    """
    Get the path to the board info file.

    Args:
        config_dir: Directory for the configuration files.

    Returns:
        The full path to the board info file.
    """
    return _file_from_path(config_dir, "board_info.json")


def save_riot_board_info(
    config_dir: Union[Path, str], board_info: Dict[str, Union[str, int]]
):
    """
    Write the board_info dict to a json file in the config_dir.

    Args:
        config_dir: Directory for the configuration files.
        board_info: Board information to save.
    """
    file_path = board_info_path(config_dir)
    with file_path.open("w") as file:
        json.dump(board_info, file, indent=2, sort_keys=True)


def load_riot_board_info(config_dir: Union[Path, str]) -> List[str]:
    """
    Load and return the riot board info from a json file in the config_dir.

    Args:
        config_dir: Directory for the configuration files.

    Returns:
        The board information.
    """
    file_path = board_info_path(config_dir)
    if not file_path.exists():
        return []

    with file_path.open() as file:
        return json.load(file)


def load_nodes(config_dir: Union[Path, str]) -> List[NmNode]:
    """
    Load and return the nodes from a json file in the config_dir.

    Args:
        config_dir: Directory for the configuration files.

    Returns:
        The nodes.
    """
    file_path = nodes_path(config_dir)
    if not file_path.exists():
        return []

    with file_path.open() as file:
        data = json.load(file)

    return [NmNode.from_dict(item) for item in data]


def save_nodes(config_dir: Union[Path, str], boards: List[NmNode]):
    """
    Write the list of boards to a json file in the config_dir.

    Args:
        config_dir (Union[Path, str]): Directory for the configuration files.
        boards (List[NmNode]): List of boards to save.
    """
    file_path = nodes_path(config_dir)
    with file_path.open("w") as file:
        data = [board.to_dict() for board in boards]
        json.dump(data, file, indent=2)


def get_lock_path(config_dir: Union[Path, str]) -> Path:
    """
    Return the lock file path constructed from config_dir.

    Args:
        config_dir (Union[Path, str]): Directory for the configuration files.

    Returns:
        Path: The full path to the lock file.
    """
    config_dir = Path(config_dir).expanduser() / "locks"
    config_dir.mkdir(parents=True, exist_ok=True)
    return config_dir


def config_arg(parser: argparse.ArgumentParser):
    """
    Add a configuration argument to the provided parser.

    Args:
        parser (argparse.ArgumentParser): ArgumentParser object.
    """
    nm_config_dir = os.environ.get("NM_CONFIG_DIR", "~/.config/inet_nm")
    parser.add_argument(
        "-c",
        "--config",
        default=nm_config_dir,
        help="Path to the config dir, defaults to NM_CONFIG_DIR or "
        "~/.config/inet_nm if NM_CONFIG_DIR is not set",
    )



"""This module contains the data types used by the inet-nm module."""
import hashlib
from dataclasses import dataclass
from typing import Any, Dict, List, Optional


class DictSerializable:
    """Mixin class to provide to_dict and from_dict methods."""

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the instance to a dictionary.

        Returns:
            Dictionary representation of the instance.
        """
        return self.__dict__

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "DictSerializable":
        """
        Create an instance from a dictionary.

        Args:
            data: Dictionary with data to populate the instance.

        Returns:
            An instance of the class that inherits this mixin.
        """
        return cls(**data)


@dataclass
class NmNode(DictSerializable):
    """
    A representation of an embedded dev board connected via USB.

    Attributes:
        serial: Serial number of the device.
        vendor_id: Vendor ID of the device.
        product_id: Product ID of the device.
        vendor: Name of the vendor.
        driver: Name of the driver.
        model: Model name of the device.
        uid: Unique ID of the device
        board: Name of the board.
        features_provided: List of features provided by the device.
    """

    serial: str
    vendor_id: str
    product_id: str
    vendor: str
    driver: str
    model: Optional[str] = None
    uid: Optional[str] = None
    board: Optional[str] = None
    features_provided: Optional[List[str]] = None

    def __post_init__(self):
        """Initialize additional attributes after instantiation."""
        self.features_provided = self.features_provided or []
        self.uid = self.uid or NmNode.calculate_uid(
            self.product_id, self.vendor_id, self.serial
        )

    @staticmethod
    def calculate_uid(product_id: str, vendor_id: str, serial: str) -> str:
        """
        Calculate a unique ID for a device.

        Args:
            product_id: Product ID of the device.
            vendor_id: Vendor ID of the device.
            serial: Serial number of the device.

        Returns:
            Calculated unique ID.
        """
        unique_key = f"{product_id}-{vendor_id}-{serial}"
        return hashlib.md5(unique_key.encode()).hexdigest()


@dataclass
class NodeEnv(DictSerializable):
    """
    Environment variables for a node.

    Attributes:
        NM_IDX: Index of the node.
        NM_UID: Unique ID of the node.
        NM_SERIAL: Serial number of the node.
        NM_BOARD: Board of the node.
        NM_PORT: Port of the node.
    """

    NM_IDX: int
    NM_UID: str
    NM_SERIAL: str
    NM_BOARD: str
    NM_PORT: str

    def to_dict(self) -> Dict[str, str]:
        """
        Convert the instance to a dictionary.

        Returns:
            Dict: Dictionary representation of the instance.

        """
        return {k: str(v) for k, v in self.__dict__.items()}



"""This module contains functions for locking nodes."""
import tempfile
from pathlib import Path
from typing import List

from inet_nm._helpers import nm_print
from inet_nm.data_types import NmNode


def locks_dir() -> Path:
    """
    Get the directory for lock files.

    If the directory does not exist, create it.

    Returns:
        The path to the lock files directory.
    """
    path = Path(tempfile.gettempdir(), "inet_nm", "locks")
    path.mkdir(parents=True, exist_ok=True)
    return path


def get_locked_uids() -> List[str]:
    """
    Get the list of UIDs of currently locked nodes.

    Returns:
        A sorted list of UIDs of locked nodes.
    """
    uids = [lock_file.stem for lock_file in locks_dir().glob("*.lock")]
    return sorted(uids)


def get_lock_path(node: NmNode) -> Path:
    """
    Get the path to the lock file for a given node.

    Args:
        node: The node for which to get the lock file path.

    Returns:
        The path to the lock file for the node.
    """
    return locks_dir() / f"{node.uid}.lock"


def release_all_locks():
    """Release all locks by deleting all lock files."""
    for lock_file in locks_dir().glob("*"):
        lock_file.unlink()


def cli_release_all_locks():
    """Release all locks by deleting all lock files and print the process."""
    for lock_file in locks_dir().glob("*"):
        nm_print(f"Removing lock file {lock_file}")
        lock_file.unlink()



from unittest.mock import patch

import pytest

from inet_nm.check import (
    check_nodes,
    eval_features,
    filter_nodes,
    get_all_features,
    remove_used_nodes,
    skip_duplicate_boards,
)
from inet_nm.data_types import NmNode


@pytest.fixture
def dummy_nodes():
    """Return a list of dummy nodes."""
    node1 = NmNode(
        serial="1",
        vendor="vendor1",
        product_id="product1",
        vendor_id="vendor_id1",
        model="model1",
        driver="driver1",
        features_provided=["feature1", "feature2"],
        board="board1",
    )
    node2 = NmNode(
        serial="2",
        vendor="vendor2",
        product_id="product2",
        vendor_id="vendor_id2",
        model="model2",
        driver="driver2",
        features_provided=["feature1"],
        board="board2",
    )
    node3 = NmNode(
        serial="3",
        vendor="vendor3",
        product_id="product3",
        vendor_id="vendor_id3",
        model="model3",
        driver="driver3",
        features_provided=["feature2", "feature3"],
        board="board3",
    )
    node4 = NmNode(
        serial="4",
        vendor="vendor4",
        product_id="product4",
        vendor_id="vendor_id4",
        model="model4",
        driver="driver4",
        features_provided=["feature2", "feature4"],
        board="board3",
    )
    return [node1, node2, node3, node4]


@pytest.fixture
def locked_nodes(dummy_nodes):
    """Return a list of locked nodes."""
    return [dummy_nodes[0].uid, dummy_nodes[1].uid]


def test_get_all_features(dummy_nodes):
    """Ensure all features are captured from the dummy nodes."""
    features = get_all_features(dummy_nodes)
    assert features == set(["feature1", "feature2", "feature3", "feature4"])


def test_filter_nodes(dummy_nodes):
    """Ensure nodes are filtered correctly using feature filter."""
    nodes = filter_nodes(dummy_nodes, ["feature1"])
    assert len(nodes) == 2
    assert nodes[0].serial == "1"
    assert nodes[1].serial == "2"


def test_skip_duplicate_boards(dummy_nodes):
    """Ensure nodes with duplicate boards are skipped."""
    nodes = skip_duplicate_boards(dummy_nodes)
    assert len(nodes) == 3
    assert nodes[0].serial == "1"
    assert nodes[1].serial == "2"
    assert nodes[2].serial == "3"


@patch("inet_nm.check.get_connected_uids")
def test_check_nodes(mock_get_connected_uids, dummy_nodes):
    """Mock the connected nodes are being captured"""
    mock_get_connected_uids.return_value = [
        dummy_nodes[0].uid,
        dummy_nodes[1].uid,
        dummy_nodes[2].uid,
    ]
    nodes = check_nodes(dummy_nodes)
    assert len(nodes) == 3


def test_remove_used_nodes(dummy_nodes, locked_nodes):
    """Ensure locked nodes are removed from the list of nodes."""
    nodes = remove_used_nodes(dummy_nodes, locked_nodes)
    assert len(nodes) == 2
    assert nodes[0].serial == "3"
    assert nodes[1].serial == "4"


def test_eval_features(dummy_nodes):
    """Ensure nodes are filtered correctly using eval feature."""
    features = get_all_features(dummy_nodes)
    nodes = eval_features(
        dummy_nodes, features, "(feature3 and feature2) or (feature1 and not feature2)"
    )
    assert len(nodes) == 2
    assert nodes[0].serial == "2"
    assert nodes[1].serial == "3"



from inet_nm.data_types import NmNode, NodeEnv


def test_nm_node():
    """Test the calculate_uid for the NmNode class."""
    uid1 = NmNode.calculate_uid("123", "456", "789")
    uid2 = NmNode.calculate_uid("321", "654", "987")
    assert uid1 == "856f4f9c3c084d978ec7ea0ad7d4cadf"
    assert uid2 == "410ddb1b4c8f37759162c9849b85c9de"
    assert uid1 != uid2

    # Test NmNode.to_dict and NmNode.from_dict
    data = {
        "serial": "serial",
        "vendor_id": "vendor_id",
        "product_id": "product_id",
        "vendor": "vendor",
        "model": "model",
        "driver": "driver",
    }
    node = NmNode.from_dict(data)

    # Test a subset of items to allow for additional attributes in the future
    for key in data:
        assert node.to_dict()[key] == data[key]


def test_node_env():
    """Test the NodeEnv data type."""
    env = NodeEnv(
        NM_IDX=1, NM_UID="uid", NM_SERIAL="serial", NM_BOARD="board", NM_PORT="port"
    )
    assert env.to_dict() == {
        "NM_IDX": "1",
        "NM_UID": "uid",
        "NM_SERIAL": "serial",
        "NM_BOARD": "board",
        "NM_PORT": "port",
    }



"""This module commissions USB boards."""
import argparse
import sys
from typing import List, Optional

import pyudev

import inet_nm.config as cfg
from inet_nm._helpers import nm_prompt_choice, nm_prompt_confirm, nm_prompt_input
from inet_nm.data_types import NmNode


class TtyNotPresent(Exception):
    """Exception to be raised when a TTY device is not found for a given NmNode."""


def get_devices_from_tty(saved_nodes: Optional[List[NmNode]] = None) -> List[NmNode]:
    """
    Retrieve connected TTY devices as a list of NmNode objects.

    Args:
        saved_nodes: List of previously saved NmNode objects.

    Returns:
        List of NmNode objects representing connected TTY devices.

    Raises:
        TtyNotPresent: If a TTY device could not be found for a given NmNode.
    """
    saved_nodes = saved_nodes or []
    context = pyudev.Context()
    nodes = []
    for device in context.list_devices(subsystem="tty"):
        if "ACM" in device.device_node or "USB" in device.device_node:
            usb_device = device.find_parent("usb", "usb_device")
            nm_node = NmNode(
                vendor_id=usb_device.get("ID_VENDOR_ID"),
                product_id=usb_device.get("ID_MODEL_ID"),
                serial=usb_device.get("ID_SERIAL_SHORT"),
                vendor=usb_device.get("ID_VENDOR_FROM_DATABASE"),
                model=usb_device.get("ID_MODEL_FROM_DATABASE"),
                driver=usb_device.get("DRIVER"),
            )

            if nm_node.uid in [node.uid for node in saved_nodes]:
                continue
            nodes.append(nm_node)
    return nodes


def get_tty_from_nm_node(nm_node: NmNode) -> str:
    """
    Retrieve the TTY device node string for a given NmNode.

    Args:
        nm_node: An NmNode object.

    Returns:
        The TTY device node string.

    Raises:
        TtyNotPresent: If a TTY device could not be found for the given NmNode.
    """
    context = pyudev.Context()
    vendor_id = nm_node.vendor_id
    model_id = nm_node.product_id
    serial_short = nm_node.serial

    for device in context.list_devices(subsystem="tty"):
        parent = device.find_parent("usb", "usb_device")
        if parent is None:
            continue
        if (
            parent.get("ID_VENDOR_ID") == vendor_id
            and parent.get("ID_MODEL_ID") == model_id
            and parent.get("ID_SERIAL_SHORT") == serial_short
        ):
            return device.device_node

    raise TtyNotPresent(f"Could not find tty device for {nm_node}")


def select_available_node(nodes: List[NmNode]) -> NmNode:
    """
    Prompt the user to select an available node from a list of NmNode objects.

    Args:
        nodes: List of available NmNode objects.

    Returns:
        The selected NmNode.
    """

    def _nice_node(node: NmNode):
        msg = f"{get_tty_from_nm_node(node)} {node.vendor}"
        if node.model:
            msg += f" {node.model}"
        msg += f" {node.serial}"
        return msg

    return nm_prompt_choice("Select the node", nodes, _nice_node)


def add_node_to_nodes(nodes: List[NmNode], node: NmNode) -> List[NmNode]:
    """
    Add a new NmNode to a list of existing NmNode objects.

    Replaces any existing NmNode with the same UID.

    Args:
        nodes: List of existing NmNode objects.
        node: New NmNode to be added.

    Returns:
        Updated list of NmNode objects.
    """
    nodes = (
        [node if n.uid == node.uid else n for n in nodes]
        if any(n.uid == node.uid for n in nodes)
        else nodes + [node]
    )
    return nodes


def select_board(board_names: List[str], node: NmNode) -> str:
    """
    Prompts the user to select a board for a given NmNode.

    Args:
        board_names: List of available board names.
        node: NmNode for which to select a board.

    Returns:
        The selected board name.
    """
    msg = f"Select board name for {node.vendor}"
    if node.model:
        msg += f" {node.model}"
    while True:
        board = nm_prompt_input(msg, board_names)
        if board in board_names:
            return board
        if nm_prompt_confirm(
            f"Board {board} not in board list, continue?", default=False
        ):
            return board


def commission():
    """CLI to commission a USB board."""
    parser = argparse.ArgumentParser(description="Commission USB boards")
    cfg.config_arg(parser)
    parser.add_argument("-b", "--board", help="Name of the board to commission")
    parser.add_argument(
        "-n", "--no-cache", help="Ignore the cache", action="store_true"
    )

    args = parser.parse_args()

    saved_nodes = cfg.load_nodes(args.config)
    nm_nodes = (
        get_devices_from_tty() if args.no_cache else get_devices_from_tty(saved_nodes)
    )
    print(f"Found {len(saved_nodes)} saved nodes in {args.config}")

    try:
        selected_node = select_available_node(nm_nodes)
    except ValueError:
        print("No available nodes found")
        return
    binfo = cfg.load_riot_board_info(args.config)
    selected_node.board = select_board(list(binfo.keys()), selected_node)
    if selected_node.board in binfo:
        selected_node.features_provided = binfo[selected_node.board]

    nodes = add_node_to_nodes(saved_nodes, selected_node)
    cfg.save_nodes(args.config, nodes)
    print(f"Updated {cfg.nodes_path(args.config)}")


def update_commissioned():
    """Go through the list of commissioned nodes and update features."""
    parser = argparse.ArgumentParser(description="Update commissioned features")
    cfg.config_arg(parser)
    args = parser.parse_args()
    binfo = cfg.load_riot_board_info(args.config)
    nodes = cfg.load_nodes(args.config)
    for node in nodes:
        if node.board in binfo:
            node.features_provided = binfo[node.board]
    cfg.save_nodes(args.config, nodes)
    print(f"Updated {cfg.nodes_path(args.config)}")


def cli_tty_from_uid():
    """CLI to get the TTY device node string for a given NmNode UID."""
    parser = argparse.ArgumentParser(
        description="Get the TTY device node string for a given NmNode UID."
    )
    cfg.config_arg(parser)
    parser.add_argument("uid", help="Node UID string.")
    args = parser.parse_args()
    nodes = cfg.load_nodes(args.config)
    try:
        node = next(node for node in nodes if node.uid == args.uid)
    except StopIteration:
        print(f"Could not find node with UID {args.uid}")
        return
    try:
        print(get_tty_from_nm_node(node))
    except TtyNotPresent:
        print(f"Could not find TTY device for {node.board} with SN {node.serial}")
        sys.exit(1)
    sys.exit(0)
